unit PixelTrans;

// TPixelTransformer (c) 2010 Piotr Daek

interface

uses Windows, Graphics, Dialogs, Sysutils;

type
  TTriByte = array[0..2] of byte;
  PTriByte = ^TTriByte;

  EPFError = Exception;

  TPixelTransformer = class(TObject)
  private
	  FColorTableOpt: array[0..255] of TTribyte;
	  FSrcBmp: TBitmap;
	  FSrcPf: TPixelFormat;
	  FSrcWidth: integer;
	  FDstBuf: PByte;
	  FLastCvtLine: integer;
  public
	  constructor Create(SrcBmp: TBitmap);
	  destructor Destroy; override;
	  function Convert(LineNum: integer): Pointer;
  end;

implementation

constructor TPixelTransformer.Create(SrcBmp: TBitmap);
var
 ColorTableOrg: array[0..255] of TRGBQuad;
 a: integer;

	function CopyPalettes(Colors: integer): boolean;
	var
	 a: integer;
	 P: PByte;
	begin;
	result:=Windows.GetPaletteEntries(SrcBmp.Palette, 0, Colors, ColorTableOrg[0])=Cardinal(Colors);
	if result then
		begin;
		p:=@FColorTableOpt[0][0];
		for a:=0 to Colors-1 do
			begin;
			p^:=ColorTableOrg[a].rgbRed;
			inc(p);
			p^:=ColorTableOrg[a].rgbGreen;
			inc(p);
			p^:=ColorTableOrg[a].rgbBlue;
			inc(p);
			end;
		end
	else
		ShowMessage(SysErrorMessage(GetLastError));
	end;

begin;
if SrcBmp=nil then
	raise EPFError.Create('Source bitmap cannot be nil.');
FSrcPf:=SrcBmp.PixelFormat;
FSrcBmp:=SrcBmp;
FSrcWidth:=SrcBmp.Width;
FLastCvtLine:=-1;
if FSrcPF=pfDevice then
	begin;
	a:=GetDC(0);
	case GetDeviceCaps(a,BITSPIXEL) of
		1: FSrcPF:=pf1bit;
		4: FSrcPF:=pf4bit;
		8: FSrcPF:=pf8bit;
		15: FSrcPF:=pf15bit;
		16: FSrcPF:=pf16bit;
		24: FSrcPF:=pf24bit;
		32: FSrcPF:=pf32bit;
		end;
	ReleaseDC(0,a);
	end;

case FSrcPf of
	pf1bit:	CopyPalettes(2);
	pf4bit:	CopyPalettes(16);
	pf8bit:	CopyPalettes(256);
	pf15bit: 	raise EPFError.Create('pf15bit bitmaps are not supported.');
	pf16bit: 	raise EPFError.Create('pf16bit bitmaps are not supported.');
	end;
if FSrcPf<>pf24bit then
	begin;
	if FSrcPF=pf32bit then
		Getmem(FDstBuf, 3*FSrcWidth+3) // extra space for fast conversion
	else
		GetMem(FDstBuf, 3*FSrcWidth);
	end
else
	FDstBuf:=nil;
end;

destructor TPixelTransformer.Destroy;
begin;
if (FSrcPf<>pf24bit)and(FDstBuf<>nil) then
	Freemem(FDstBuf);
Inherited;
end;

function TPixelTransformer.Convert(LineNum: integer): Pointer;
var
 Src: PByte;
 Dst, DstMax: PTriByte;
 x, xmax: integer;
 sb: byte;

	procedure Convert32RowTo24Bit(Src, Dst: pointer; SrcMax: integer);
	asm
		test ecx, ecx
		jz @quit
		add ecx, eax
		push ebx
		mov ebx, eax
@loop:
    // note that we write in 4-byte packets, so without extra space we could
    // overwrite anything that lies right after destination buffer.
		mov eax, [ebx]
		mov [edx], eax
		add edx, 3
		add ebx, 4
		cmp ebx, ecx
		jne @loop
		pop ebx;
@quit:
	end;

begin;
// no range checking - FSrcBmp.Scanline will raise appriopriate exception if
// we're out of bounds.
if FLastCvtLine=LineNum then
	begin;
	result:=FDstBuf;
	exit;
	end;
FLastCvtLine:=LineNum;
src:=FSrcBmp.Scanline[Linenum];
if FSrcPf=pf24bit then
	begin;
	FDstBuf:=src;
	result:=src;
	exit;
	end;
xmax:=FSrcWidth*3;
result:=FDstBuf;
Dst:=Pointer(FDstBuf);
DstMax:=Pointer(longint(Dst)+xmax);
case FSrcPf of
	pf1bit:	begin;
			while true do
				begin;
				sb:=src^;
				if cardinal(Dst)+24<cardinal(Dstmax) then
					for x:=7 downto 0 do
						begin;
						Dst^:=FColorTableOpt[(sb shr x) and 1];
						inc(Dst);
						end
				else
					begin;
					for x:=7 downto 0 do
						begin;
						Dst^:=FColorTableOpt[(sb shr x) and 1];
						inc(Dst);
						if cardinal(Dst)>=cardinal(Dstmax) then
							break;
						end;
					break;
					end;
				inc(src);
				end;
			end;
	pf4bit:	begin;
			while cardinal(Dst)<cardinal(Dstmax) do
				begin;
				sb:=src^;
				Dst^:=FColorTableOpt[(sb and $F0) shr 4];
				inc(Dst);
				if cardinal(Dst)>=cardinal(Dstmax) then
					break;
				Dst^:=FColorTableOpt[sb and $0F];
				inc(src);
				inc(Dst);
				end;
			end;
	pf8bit:	begin;
			while cardinal(Dst)<cardinal(Dstmax) do
				begin;
				Dst^:=FColorTableOpt[Src^];
				inc(src);
				inc(Dst);
				end;
			end;
	pf32Bit:	Convert32RowTo24bit(Src, Dst, FSrcWidth*4);
	else		result:=Src;
	end;
end;


end.

